@@ -57,18 +57,22 @@ class DB(object):
     class VariableTypeError(Exception): pass
     class VariableUndefinedError(Exception): pass
     class FileNotInDB(Exception): pass
-    def __init__(self, callback=None, filename=None):
+    def __init__(self, callback=None, profile=None):
         """
         callback is called to display progress when scanning lesson files.
+        profile None is the default profile stored in app_data(),
+        if profile is PROFILENAME, this is the profile stored in
+        app_data()/profiles/PROFILENAME
         """
-        if filename:
-            statistics_filename = filename
-        else:
-            try:
-                if testsuite_is_running:
-                    statistics_filename = ":memory:"
-            except NameError:
-                statistics_filename = self.get_statistics_filename()
+        self.m_profile = profile
+        try:
+            if testsuite_is_running:
+                statistics_filename = ":memory:"
+        except NameError:
+            statistics_filename = self.get_statistics_filename()
+            head, tail = os.path.split(statistics_filename)
+            if not os.path.exists(head):
+                os.makedirs(head)
         self.must_recreate = not os.path.exists(statistics_filename)
         try:
             if testsuite_is_running:
@@ -77,8 +81,12 @@ class DB(object):
             pass
         self.conn = sqlite3.connect(statistics_filename)
         self.setup_tables()
-        if self.must_recreate:
+        # We don't read old data if we are running another profile than
+        # the default one, since there are no old format statistics saved
+        # in the profiles
+        if self.must_recreate and not profile:
             self.read_old_data(callback)
+        self.upgrade_to_version_2()
     def insert_file(self, filename):
         assert lessonfile.is_uri(filename) or os.path.isabs(filename)
         self.conn.execute("insert into lessonfiles "
@@ -87,8 +95,18 @@ class DB(object):
                 (filename, hash_of_lessonfile(filename)))
         return self.get_fileid(filename)
     @staticmethod
-    def get_statistics_filename():
+    def get_noprofile_statistics_filename():
         return os.path.join(filesystem.app_data(), "statistics.sqlite")
+    def get_statistics_filename(self):
+        if self.m_profile:
+            return os.path.join(filesystem.app_data(), "profiles", self.m_profile, "statistics.sqlite")
+        else:
+            return self.get_noprofile_statistics_filename()
+    def reset_database(self):
+        self.conn.close()
+        os.remove(self.get_statistics_filename())
+        self.conn = sqlite3.connect(self.get_statistics_filename())
+        self.setup_tables()
     def setup_tables(self):
         """
         Create needed tables. Drop and recreate the database if it was
@@ -119,12 +137,14 @@ class DB(object):
             ")")
         self.conn.execute('''create table if not exists sessions
             (fileid int, timestamp int, answerkey text, guessed text, count int, unique (fileid, timestamp, answerkey, guessed) )''')
-        self.conn.execute('''create table if not exists tests
-            (fileid int, timestamp int, answerkey text, guessed text, count int, unique (fileid, timestamp, answerkey, guessed) )''')
         self.conn.execute('''create table if not exists variables
             (variable_name text primary key not null,
              type int not null,
              value text not null)''')
+        self.conn.execute("create table if not exists sessioninfo "
+            "(fileid int, timestamp int, sessiontype int, "
+            "unique (fileid, timestamp))")
+        self.set_variable("database_version", 2)
     def get_fileid(self, filename):
         assert lessonfile.is_uri(filename) or os.path.isabs(filename), filename
         row = self.conn.execute(
@@ -180,7 +200,7 @@ class DB(object):
             return unpickler.load() 
 
         imp, uuid_to_filename = self.get_uuid_to_filename_mapping(callback)
-        def read_session_data(datadir, tablename, counter):
+        def read_session_data(datadir, is_tests, counter):
             if not os.path.isdir(datadir):
                 logging.debug("read_session_data: '%s' does not exist. Returning" % datadir)
                 return
@@ -247,17 +267,22 @@ class DB(object):
                     f.close()
                     for correct in session.keys():
                         for guess in session[correct]:
-                            self.conn.execute("insert into %s "
+                            self.conn.execute("insert into sessions "
                                 "(fileid, timestamp, answerkey, guessed, count) "
-                                "values(?, ?, ?, ?, ?)" % tablename,
+                                "values(?, ?, ?, ?, ?)",
                             (fileid, timestamp, unicode(correct), unicode(guess), session[correct][guess]))
+                    self.conn.execute("insert into sessioninfo "
+                        "(fileid, timestamp, sessiontype) "
+                        "values(?, ?, ?)",
+                        (fileid, timestamp, int(bool(is_tests))))
                     counter += 1
                     callback(_("Files read: %i") % counter)
-                if tablename == 'tests':
-                    row = self.conn.execute("select timestamp from tests "
-                            "where fileid=? order by -timestamp",
-                            (fileid,)).fetchone()
-                    if row:
+                if is_tests:
+                    timestamp = self.conn.execute("select timestamp from sessioninfo "
+                        "where fileid=? and sessiontype=1 "
+                        "order by -timestamp ",
+                        (fileid,)).fetchone()[0]
+                    if timestamp:
                         if lessonfile.infocache.get(lessonfilename, "module") in ('melodicinterval', 'harmonicinterval', 'singinterval'):
                             parserclass = lessonfile.IntervalsLessonfile
                         else:
@@ -265,44 +290,114 @@ class DB(object):
                         p = parserclass()
                         p.parse_file(lessonfilename)
                         self.cache_new_test_result(lessonfilename,
-                            row[0],
+                            timestamp,
                             p.get_test_requirement(),
                             p.get_test_num_questions())
             return imp
         imp = read_session_data(
                 os.path.join(filesystem.app_data(), 'statistics'),
-                "sessions", imp)
+                False, imp)
         imp = read_session_data(
                 os.path.join(filesystem.app_data(), 'testresults'),
-                "tests", imp)
+                True, imp)
+        self.set_variable("database_version", 2)
         logging.info("imported statistics for %s lessonfiles" % imp)
         self.conn.commit()
+    def upgrade_to_version_2(self):
+        """
+        After releaseing 3.16.0 I found that having separate tests and
+        sessions table was a bad idea since we may want to add different
+        tests later. Also tests and practise sessions in one table make
+        it even easier to see which questions needs most practise.
+        """
+        try:
+            if self.get_variable("database_version") >= 2:
+                return
+        except self.VariableUndefinedError:
+            pass
+        # Insert all tests into sessioninfo
+        for fileid, timestamp in self.conn.execute("select distinct fileid, timestamp from tests"):
+            # This should always succeed since we have just created sessioninfo.
+            # But let us try: just to be safe.
+            try:
+                self.conn.execute("insert into sessioninfo "
+                    "(fileid, timestamp, sessiontype) "
+                    "values (?, ?, ?)",
+                    (fileid, timestamp, 1))
+            except sqlite3.IntegrityError, e:
+                logging.error(u"%s, %s, %s" % (e, fileid, timestamp))
+        # Insert all practise sessions into sessioninfo
+        for fileid, timestamp in self.conn.execute("select distinct fileid, timestamp from sessions"):
+            # This can fail because 3.14.11 and earlier had a bug where the same
+            # session was create in both statistics and testresult.
+            try:
+                self.conn.execute("insert into sessioninfo "
+                    "(fileid, timestamp, sessiontype) "
+                    "values (?, ?, ?)",
+                    (fileid, timestamp, 0))
+            except sqlite3.IntegrityError, e:
+                logging.error(u"%s, %s, %s" % (e, fileid, timestamp))
+        # Copy all tests from the obsolete "tests" table to "sessions"
+        for a, b, c, d in self.conn.execute("select fileid, timestamp, answerkey, guessed from tests"):
+            # This can fail because 3.14.11 and earlier had a bug where the same
+            # session was create in both statistics and testresult.
+            try:
+                self.conn.execute("insert into sessions "
+                    "(fileid, timestamp, answerkey, guessed) "
+                    "values (?, ?, ?, ?)",
+                    (a, b, c, d))
+            except sqlite3.IntegrityError, e:
+                logging.error(u"%s, %s, %s" % (e, fileid, timestamp))
+        for fileid, filename in self.conn.execute(
+                "select sessioninfo.fileid, lessonfiles.filename "
+                "from sessioninfo, lessonfiles "
+                "where sessioninfo.fileid=lessonfiles.fileid "
+                "      and sessioninfo.sessiontype<>0 "):
+            for timestamp in self.conn.execute(
+                "select timestamp from sessioninfo where fileid=? "
+                "order by -timestamp",
+                (fileid,)).fetchone():
+                module = lessonfile.infocache.get(filename, "module")
+                if module == 'melodicinterval':
+                    parserclass = lessonfile.IntervalsLessonfile
+                else:
+                    parserclass = None
+                if parserclass:
+                    p = parserclass()
+                    p.parse_file(filename)
+                    self.cache_new_test_result(filename, timestamp,
+                        p.get_test_requirement(), p.get_test_num_questions())
+        self.set_variable("database_version", 2)
     def cache_new_test_result(self, filename, timestamp, 
                               required, num_questions):
         """
-        Search the "tests" table and find the latests test results for
-        the file. Save the test result in the "lessonfiles" table for
-        faster access later.
+        Save the test result in the "lessonfiles" table for faster
+        access later.
         """
         logging.debug("cache_new_test_result(%s, %s, %s, %s)"
             % (filename, timestamp, required, num_questions))
         fileid = self.get_fileid(filename)
-        test_result = self.get_correct_count(fileid, timestamp) * 1.0 / num_questions
+        count_correct = self.conn.execute("select sum(count) from sessions where fileid=? and timestamp=? and answerkey=guessed", (fileid, timestamp)).fetchone()[0]
+        if not count_correct:
+            count_correct = 0
+        count_total = self.conn.execute("select sum(count) from sessions where fileid=? and timestamp=?", (fileid, timestamp)).fetchone()[0]
+        # count_total is 0 if the user click cancel before answering
+        # any questions. If so, we returns without saving anything. I don't
+        # see the point in saving tests where no questions where answered.
+        if not count_total:
+            return
+        if count_total < num_questions:
+            count_total = num_questions
+        if count_total:
+            test_result = count_correct * 1.0 / count_total
+        else:
+            test_result = 0.0
         self.conn.execute("update lessonfiles "
                         "set test_result=?, test_passed=? where fileid=?",
             (test_result,
              test_result >= required,
              fileid,))
         self.conn.commit()
-    def get_correct_count(self, fileid, timestamp):
-        """
-        Return the number of correct answers the test of fileid with the
-        specified timestamp.
-        """
-        correct_count = self.conn.execute("select sum(count) from tests where fileid=? and timestamp=? and answerkey=guessed", (fileid, timestamp)).fetchone()[0]
-        if not correct_count:
-            correct_count = 0
-        return correct_count
     def get_test_status(self, filename):
         """
         Return a tuple saying if the test was passed or not, and the result:
@@ -350,17 +445,40 @@ class DB(object):
                     replaces = [replaces]
                 if not hashvalue in replaces:
                     cursor.execute("delete from sessions where fileid=?", (fileid,))
-                    cursor.execute("delete from tests where fileid=?", (fileid,))
+                    cursor.execute("delete from sessioninfo where fileid=?", (fileid,)) 
                     cursor.execute("update lessonfiles "
                         "set hash=?, test_passed=?, test_result=0.0 where fileid=?",
                         (cur_lessonfile_hash_value, None, fileid))
                     self.conn.commit()
-    def _recent(self, count, tablename):
-        assert tablename in ('sessions', 'tests')
+    def get_statistics_info(self):
+        """
+        Return information about the data installed.
+        Return a tuple (number of different exercises practised,
+                        number of times we have practised,
+                        number of times we have taken a test)
+        """
+        session_count = self.conn.execute('select count(fileid) '
+            'from sessioninfo '
+            'where sessiontype=0').fetchone()[0]
+        test_count = self.conn.execute('select count(fileid) '
+            'from sessioninfo '
+            'where sessiontype=1').fetchone()[0]
+        different_ex = len(self.conn.execute('select fileid from sessioninfo group by fileid').fetchall())
+        return {'exercises': different_ex,
+                'practise_count': session_count,
+                'test_count': test_count}
+
+    def _recent(self, count, sessiontype):
+        """
+        sessiontype 0 == normal statistics
+        sessiontype 1 == test results
+        """
         filenames = []
         for fileid, timestamp in self.conn.execute(
-                'select fileid, timestamp from %s '
-                'order by -timestamp' % tablename):
+                'select fileid, timestamp from sessioninfo '
+                'where sessiontype=? '
+                'order by -timestamp',
+                (sessiontype,)):
             filename = self.conn.execute('select filename from lessonfiles where fileid=?', (fileid,)).fetchone()[0]
             if filename not in filenames:
                 filenames.append(filename)
@@ -368,9 +486,9 @@ class DB(object):
                 break
         return filenames
     def recent(self, count):
-        return self._recent(count, 'sessions')
+        return self._recent(count, 0)
     def recent_tests(self, count):
-        return self._recent(count, 'tests')
+        return self._recent(count, 1)
     def set_variable(self, name, value):
         """
         raise DB.VariableTypeError if the type we set is different from
@@ -438,10 +556,16 @@ class AbstractStatistics(object):
         except DB.FileNotInDB:
             return []
         if all_keys:
-            c = solfege.db.conn.execute("select distinct(answerkey) from sessions where fileid=?", (fileid,))
+            c = set()
+            for colname in "answerkey", "guessed":
+                c1 = list(solfege.db.conn.execute("select distinct(%s) from sessions where fileid=?" % colname, (fileid,)))
+                if c1:
+                    [c.add(x[0]) for x in c1]
+            c = list(c)
         else:
             c = solfege.db.conn.execute("select distinct(answerkey) from sessions where fileid=? and answerkey=guessed", (fileid,))
-        v = [self.int_if_int(x[0]) for x in c]
+            c = [x[0] for x in list(c)]
+        v = [self.int_if_int(x) for x in c]
         v.sort()
         return [unicode(x) for x in v]
     def get_statistics(self, seconds):
@@ -451,6 +575,9 @@ class AbstractStatistics(object):
         And the values of dict are new dicts where the keys are all the
         answers the user have given and the values are the number of times
         that particular answer has been given.
+        Special values of second:
+        -1: all history
+        0: statistics from the current session
         """
         try:
             fileid = solfege.db.get_fileid(self.m_t.m_P.m_filename)
@@ -485,6 +612,7 @@ class AbstractStatistics(object):
                 self.m_timestamp,
                 self.m_t.m_P.get_test_requirement(),
                 self.m_t.m_P.get_test_num_questions())
+        self.reset_session()
     def _add(self, question, answer):
         """
         Register that for the question 'question' the user answered 'answer'.
@@ -495,25 +623,34 @@ class AbstractStatistics(object):
             question = str(question)
         if isinstance(answer, tuple):
             answer = str(answer)
-        table = {True: 'tests', False: 'sessions'}[self.m_test_mode]
         cursor = solfege.db.conn.cursor()
         fileid = solfege.db.get_fileid(self.m_t.m_P.m_filename)
+        # Let us check if the session has been added to "sessioninfo"
+        # We don't add the session to the "sessioninfo" table before we have
+        # an answer to store, to avoid empty sessions because users start
+        # an exercise and the descides it was the wrong exercise.
+        if not cursor.execute("select * from sessioninfo where fileid=? and timestamp=?", (fileid, self.m_timestamp)).fetchone():
+            cursor.execute("insert into sessioninfo"
+                    "(fileid, timestamp, sessiontype) "
+                    "values (?, ?, ?)",
+                    (fileid, self.m_timestamp, 1 if self.m_test_mode else 0))
+
         row = cursor.execute(
-                "select count from %s where fileid=? and timestamp=? "
-                "and answerkey=? and guessed=?" % table,
+                "select count from sessions where fileid=? and timestamp=? "
+                "and answerkey=? and guessed=?",
                 (fileid, self.m_timestamp, unicode(question), unicode(answer))).fetchone()
         if not row:
             cursor.execute(
-                "insert into %s "
+                "insert into sessions "
                 "(fileid, timestamp, answerkey, guessed, count) "
-                "values(?, ?, ?, ?, ?)" % table,
+                "values(?, ?, ?, ?, ?)",
                 (fileid, self.m_timestamp, 
                  unicode(question), unicode(answer), 1))
         else:
             assert cursor.fetchone() is None
             cursor.execute(
-                "update %s set count=? where "
-                "fileid=? and timestamp=? and answerkey=? and guessed=?" % table,
+                "update sessions set count=? where "
+                "fileid=? and timestamp=? and answerkey=? and guessed=?",
                 (row[0] + 1, fileid,
                  self.m_timestamp, unicode(question), unicode(answer)))
         solfege.db.conn.commit()
@@ -528,12 +665,10 @@ class AbstractStatistics(object):
         return solfege.db.last_test_result(self.m_t.m_P.m_filename)
     def get_percentage_correct(self):
         """Will return a 0 <= value <= 100.0 that say how many percent is
-        correct in this session
+        correct in this session.
         """
         fileid = solfege.db.get_fileid(self.m_t.m_P.m_filename)
-        fileid = ["None"]
         num_correct = solfege.db.conn.execute("select sum(count) from sessions where answerkey=guessed and timestamp=? and fileid=?", (self.m_timestamp, fileid)).fetchone()[0]
-        print "Num", num_correct
         num_asked = solfege.db.conn.execute("select sum(count) from sessions where timestamp=? and  fileid=?", (self.m_timestamp, fileid)).fetchone()[0]
         if not num_correct:
             num_correct = 0
@@ -582,6 +717,27 @@ class AbstractStatistics(object):
         if ret:
             return ret
         return 0
+    def iter_test_results(self):
+        """
+        Iterate test results for the associated lesson file, newest
+        results first.
+        """
+        try:
+            fileid = solfege.db.get_fileid(self.m_t.m_P.m_filename)
+        except DB.FileNotInDB:
+            return
+        for [timestamp] in solfege.db.conn.execute("select timestamp from sessioninfo where fileid=? and sessiontype=? order by -timestamp", (fileid, 1)):
+            ret = {}
+            for answerkey, guessed, count in solfege.db.conn.execute("select answerkey, guessed, count from sessions where fileid=? and timestamp=?", (fileid, timestamp)):
+                ret.setdefault(self.int_if_int(answerkey), {})
+                ret[self.int_if_int(answerkey)][self.int_if_int(guessed)] = count
+            # More necessary than one would expect because we want to handle
+            # the possibility that that ret[key1][key2] == None
+            # A user has reported that this can happen, but I don't know
+            # what could insert a None into the database.
+            f = (100.0 * sum([y for y in [ret[x].get(x, 0) for x in ret] if y is not None])
+                 / self.m_t.m_P.get_test_num_questions())
+            yield timestamp, f, ret
 
 
 class LessonStatistics(AbstractStatistics):
@@ -601,8 +757,17 @@ class LessonStatistics(AbstractStatistics):
 
 class IntervalStatistics(AbstractStatistics):
     def get_keys(self, all_keys=False):
+        # FIXME we have to check that all keys are integers, and filter
+        # out those that are not. I don't know how, but I ended up with
+        # the key None inserted into the database.
+        def isinteger(s):
+            try:
+                int(s)
+                return True
+            except ValueError:
+                return False
         v = AbstractStatistics.get_keys(self, all_keys)
-        return [int(x) for x in v]
+        return [int(x) for x in v if isinteger(x)]
     def key_to_pretty_name(self, key):
         return utils.int_to_intervalname(int(key), 1, 1)
 
